# -*- coding:cp1251 -*-

import nltk, os, math
import cPickle
import re

numberUnigrams = 2
numberBigrams = 7

pathToTest = "./data/"
pathToTrain = "./data/"
pathToAnswers = "./answers/"
pathToModel = "./models/ngram.pkl"

# all methods for text processing
class TextProcessor():
    def __init__(self, text):
        words = nltk.PunktWordTokenizer().tokenize(text)
        words = filter(lambda x: len(x) > 3, words)  # filter 1-2 symbols words
        stemmer = nltk.stem.snowball.RussianStemmer() 
        self.words = map(stemmer.stem, words)  # apply stemmer to each word
        self.originalForm = self.saveOriginalFormOfWords(words)
    
    # save original form of word
    # here it's first occurred word of the text corresponded to the particular token
    def saveOriginalFormOfWords(self, tokenizedText):            
        originalForm = dict()
        for i in xrange(len(self.words)):
            if not(originalForm.has_key(self.words[i])):
                originalForm[self.words[i]] = tokenizedText[i]
        return originalForm  
    
    # simple word restorer
    def wordRestorer(self, word):
        return self.originalForm[word]
        
# training of algorithms
# here we only compute IDf
class Trainer():
    # model is a pair of (dict: word -> idf_score) and (number of files in training set)
    model = None
    
    def __init__ (self):
        if not(self.loadModel()):
            self.model = self.computeIDF()
            self.saveModel()
        
    def computeIDF(self):
        fileCounter = 0
        docs = dict() 
        for fileName in os.listdir(pathToTrain):
            print fileCounter,  # progress
            fileCounter += 1
                        
            f = open(pathToTrain + fileName)
            text = unicode(f.read(), "utf-8")
            f.close()
            
            # preprocess text
            tp = TextProcessor(text) 
            
            # process unigrams
            for unigram in set(tp.words):
                docs[unigram] = docs.setdefault(unigram, 0.0) + 1.0
            
            # process bigrams
            for bigram in set(NgramProcessor().getBigrams(tp.words)):
                docs[bigram] = docs.setdefault(bigram, 0.0) + 1.0
                
            
        # compute IDF
        idf = dict()
        for ngram in docs.keys():
            if docs[ngram] > 1:
                idf[ngram] = -math.log(docs[ngram] / fileCounter)
            
        return (idf, fileCounter)
    
    def saveModel(self):
        f = open(pathToModel, "wb")
        cPickle.dump(self.model, f)
        f.close()
    
    def loadModel(self):
        try:
            f = open(pathToModel)
            self.model = cPickle.load(f)
            f.close()
            return True
        except:
            return False
        
    def getModel(self):
        return self.model

# all functions for processing n-gramms
class NgramProcessor():
    def getBigrams(self, wordList):
        return zip(wordList[:-1], wordList[1:])
        
    # return sorted list of pairs (object, number of its occurrences in the list)
    # it is used for counting unigrams and bigrams in text
    def objectCounter(self, listOfObjects):
        return [(obj, listOfObjects.count(obj)) for obj in set(listOfObjects)] 
        
    # return TopN object (unigram of bigram) from list    
    def getTopN(self, listOfObjectsWithCounts, N):
        sortedListOfObjectsWithCounts = sorted(listOfObjectsWithCounts, key=lambda x:-int(x[1]))[:N]
        # return sortedListOfObjectsWithCounts
        return [item[0] for item in sortedListOfObjectsWithCounts]    

    # Assign weights to unigrams and bigrams
    # here we use TF*IDF weighting scheme
    def weightingScheme(self, listOfCountedObjects, model):
        return [(objectName, objectCount * model[0].setdefault(objectName, -math.log(1.0 / model[1]))) 
                for objectName, objectCount in listOfCountedObjects]
            
class Keywords():
    
    def __init__(self):
        # load trained model
        self.model = Trainer().getModel()
        
    def getKeywords(self, text):
        # text processor
        tp = TextProcessor(text)
        ap = tp.words[1:100]

        ng = NgramProcessor()
        bigrams = ng.getBigrams(tp.words)
        abigrams = ng.getBigrams(ap)
         
        tp.words = filter(lambda x: x in ap, tp.words)
        bigrams = filter(lambda x: x in abigrams, bigrams)
        
        unigramKeywords = map(tp.wordRestorer, #restore original forms
                              ng.getTopN(ng.weightingScheme(ng.objectCounter(tp.words), self.model), 
                                         numberUnigrams))
        bigramKeywords = [tp.wordRestorer(word[0]) + " " + tp.wordRestorer(word[1]) 
                          for word in ng.getTopN(ng.weightingScheme(ng.objectCounter(bigrams),
                                                                  self.model), numberBigrams)]
        keywords = unigramKeywords + bigramKeywords
        end2 = unicode("         ", "cp1251")
        end3 = unicode("     ", "cp1251")
        end2 = nltk.PunktWordTokenizer().tokenize(end2)
        end3 = nltk.PunktWordTokenizer().tokenize(end3)        
        keywords = filter(lambda x: x[-2:] not in set(end2) and x[-3:] not in set(end3), keywords)
        rusSeq = re.compile(unicode('[- -]+', 'cp1251'))
        keywords = filter(lambda x: rusSeq.match(x) != None and (rusSeq.match(x)).end() == len(x), keywords)
        return keywords
    
        
## QUALOTY CONTROL SECTION
#kw = Keywords()
#stemmer = nltk.stem.snowball.RussianStemmer() 
#tp = 0
#fp = 0
#fn = 0

#counter = 0
#fout = open(pathToTest + "../" + "!keywords.txt", "w")
#for fileName in os.listdir(pathToTest):   
#    counter += 1   
#    if counter < 100:
#        continue
#    print counter
#    f = open(pathToTest + fileName)
#    text = unicode(f.read(), "utf-8")
#    f.close()
#    keywords = kw.getKeywords(text)
#    for i in keywords:
#        print >> fout, i.encode('utf-8')
#    
#    f = open(pathToAnswers + fileName)
#    text = unicode(f.read(), "utf-8")
#    f.close()
#    realKeywords = nltk.LineTokenizer().tokenize(text)
#    our = set([' '.join(map(stemmer.stem, nltk.PunktWordTokenizer().tokenize(obj))) for obj in keywords])
#    their = set([' '.join(map(stemmer.stem, nltk.PunktWordTokenizer().tokenize(obj))) for obj in realKeywords])
#    tp += len(our.intersection(their))
#    fp += len(our.difference(their))
#    fn += len(their.difference(our))

#fout.close()
#p = 1.0 * tp / (tp + fp)
#r = 1.0 * tp / (tp + fn)
#F = 2 * p * r / (p + r)
#print p, r, F
